/*
 * Hornsey - Moblin Media Player.
 * Copyright © 2009 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU Lesser General Public License,
 * version 2.1, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <config.h>

#include <clutter/clutter.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <gio/gio.h>
#include <stdlib.h>
#include <string.h>

#include <dbus/dbus-protocol.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <bickley/bkl-utils.h>
#include <bickley/bkl-source-client.h>
#include <bickley/bkl-source-manager-client.h>

#include <bognor/br-iface-player.h>
#include <bognor/br-queue.h>

#define HRN_SN_SUPPORT 1
#ifdef HRN_SN_SUPPORT
#define SN_API_NOT_YET_FROZEN
#include <libsn/sn.h>
#endif

#include <clutter/x11/clutter-x11.h>

#include <clutter-gtk/clutter-gtk.h>

#include "hrn.h"
#include "hrn-state-manager.h"
#include "hrn-pin-manager.h"
#include "hrn-ui-controls.h"
#include "hrn-window.h"

#include "math.h"

#include <unique/unique.h>

#define HORNSEY_DBUS_SERVICE    "org.moblin.Hornsey"
#define HORNSEY_DBUS_PATH       "/org/moblin/Hornsey"

HrnWindow *window;

ClutterActor *hrn_preload_block;
ClutterActor *hrn_preload_content_block;
ClutterActor *hrn_ui_controls;
ClutterActor *hrn_search_entry;

ClutterActor *zoom_indicator;
ClutterActor *hrn_scrollbar;
ClutterActor *loc_indicator;
ClutterActor *hrn_queues;

ClutterActor *hrn_content_area;
ClutterActor *hrn_theatre_parent;
ClutterActor *hrn_theatre;

ClutterActor *hrn_view;

HrnQueue *master_queue; /* FIXME: Need queue manager */
ClutterActor *master_controls;
HrnPinManager *pin_manager;

HrnSource *local_source      = NULL;
GList *transient_sources = NULL; /* Contains HrnSource */

BklSourceManagerClient *manager     = NULL;
BrQueue *local_queue = NULL;

HrnStateManager *state_manager;

/* This is so we know when all the sources are loaded */
static int   source_count   = 0;
static int   source_replies = 0;

static guint controls_hide_handler_id = 0;

static UniqueResponse
hrn_message_received_cb (UniqueApp *app, gint command,
                         UniqueMessageData *message_data, guint time_,
                         gpointer user_data)
{
  gchar *uri;

  switch (command)
    {
      case UNIQUE_ACTIVATE:
        hrn_window_present (window);
        return UNIQUE_RESPONSE_OK;

      case UNIQUE_OPEN:
        uri = unique_message_data_get_text (message_data);
        if (uri && (uri[0] != '\0'))
          {
            BklItem *item;

            item = hrn_get_item_for_uri (uri);
            if (item)
              {
                hrn_set_theatre ();

                if (bkl_item_get_item_type (item) == BKL_ITEM_TYPE_IMAGE) {
                  /* We disable the slideshow so that the image will not
                     expire */
                  hrn_theatre_set_image_slideshow ((HrnTheatre *) hrn_theatre,
                                                   FALSE);
                }

                hrn_play_now (item);
              }
            else
              {
                hrn_set_theatre ();
                hrn_play_uri_now (uri);
              }
          }

        g_free (uri);
        hrn_window_present (window);

        return UNIQUE_RESPONSE_OK;

      default:
        g_print ("uh?! %i\n", command);
        break;
    }
  return UNIQUE_RESPONSE_PASSTHROUGH;
}

static void
fade_controls_sanity (ClutterAnimation *animation,
                      ClutterActor     *controls)
{
  if (!hrn_controls_can_has_fade)
    clutter_actor_set_opacity (master_controls, 0xff);
  else
    clutter_actor_hide (master_controls);
}

#if 0
static gboolean
fade_out2 (gpointer foo)
{
  if (!hrn_controls_can_has_fade)
    return FALSE;
  clutter_actor_animate (master_controls, CLUTTER_EASE_IN_CUBIC,
                         CONTROLS_OUT_DURATION3,
                         "opacity", 0,
                         "signal::completed", fade_controls_sanity,
                         master_controls,
                         NULL);
  return FALSE;
}
#endif

static gboolean
fade_out (gpointer foo, gpointer *bar)
{
  if (!hrn_controls_can_has_fade
      || master_controls == NULL ||
      clutter_actor_get_paint_visibility (master_controls) == FALSE)
    return FALSE;
  clutter_stage_hide_cursor (CLUTTER_STAGE (hrn_window_get_stage (window)));
  return FALSE;
}

static gboolean
controls_hide_handler (gpointer foo)
{
  if (!hrn_controls_can_has_fade || master_controls == NULL ||
      clutter_actor_get_paint_visibility (master_controls) == FALSE)
    {
      controls_hide_handler_id = 0;
      return FALSE;
    }
  clutter_actor_animate (master_controls, CLUTTER_LINEAR, CONTROLS_OUT_DURATION,
                         "opacity", 0,
                         "signal::completed", fade_out, NULL,
                         "signal::completed", fade_controls_sanity, NULL,
                         NULL);
  controls_hide_handler_id = 0;

  return FALSE;
}

gboolean hrn_controls_can_has_fade = FALSE;

void
hrn_controls_wake (void)
{
  if (controls_hide_handler_id)
    {
      g_source_remove (controls_hide_handler_id);
      controls_hide_handler_id = 0;
    }
  clutter_actor_show (master_controls);
  clutter_stage_show_cursor (CLUTTER_STAGE (hrn_window_get_stage (window)));

  if (!hrn_controls_can_has_fade)
    {
      return;
    }

  clutter_actor_animate (master_controls, CLUTTER_LINEAR, CONTROLS_IN_DURATION,
                         "opacity", 255,
                         NULL);
  controls_hide_handler_id =
    g_timeout_add (HRN_CONTROLS_TIMEOUT, controls_hide_handler,
                   NULL);
}

static gboolean
controls_capture_cb (ClutterActor *actor, ClutterEvent *event, gpointer data)
{
  switch (event->any.type)
    {
      case CLUTTER_LEAVE:
        if (0)
          {
            ClutterActor *actor = event->motion.source;

            g_print ("leave.. %p %p ",
                     (event->motion.source),
                     (event->crossing.related));

            if (event->crossing.related)
              g_print ("leave.. %s %s|",
                       G_OBJECT_TYPE_NAME (event->motion.source),
                       G_OBJECT_TYPE_NAME (event->crossing.related));
            while (actor)
              {
                g_print ("%s<", G_OBJECT_TYPE_NAME (actor));
                actor = clutter_actor_get_parent (actor);
              }
            g_print ("\n");
          }
        break;

      case CLUTTER_KEY_PRESS:
      case CLUTTER_KEY_RELEASE:

        if (event->key.keyval == CLUTTER_q &&
            clutter_event_get_state (event) & CLUTTER_CONTROL_MASK)
          {
            hrn_quit ();
          }

      /* some keybindings should not wake the controls */
        if (event->key.keyval == CLUTTER_Shift_L ||
            event->key.keyval == CLUTTER_Shift_R ||
            event->key.keyval == CLUTTER_space)
          return FALSE;
      case CLUTTER_MOTION:
#if 0
       /* debug code used when picking goes awry */
          {
            ClutterActor *actor = event->motion.source;

            while (actor)
              {
                g_print ("%s<", G_OBJECT_TYPE_NAME (actor));
                actor = clutter_actor_get_parent (actor);
              }
            g_print ("\n");
          }
#endif

      case CLUTTER_BUTTON_PRESS:
      case CLUTTER_BUTTON_RELEASE:
      case CLUTTER_SCROLL:
        hrn_controls_wake ();

      default:
        if (0) g_print ("%i %s %p\n", event->type,
                        G_OBJECT_TYPE_NAME (clutter_event_get_source (event)),
                        clutter_event_get_source (event));
        break;
    }
  return FALSE;
}

static void
zoom_changed_cb (HrnUiControls *controls,
                 HrnZoomLevel   old_level,
                 HrnZoomLevel   new_level,
                 gpointer       userdata)
{
  double zl;

  g_print ("Zoom changed %d -> %d\n", old_level, new_level);

  if (new_level == HRN_ZOOM_LEVEL_THEATRE) {
    hrn_theatre_show ((HrnTheatre *) hrn_theatre);
    hrn_ui_controls_hide (controls);
#if 0
    clutter_actor_animate (hrn_content_area, CLUTTER_EASE_IN_OUT_CUBIC,
                           SIDEBAR_OUT_DURATION,
                           "depth", 0.0,
                           "x", 0.0,
                           NULL);
#endif

    clutter_stage_set_key_focus (CLUTTER_STAGE (hrn_window_get_stage (window)),
                                 (ClutterActor *) hrn_theatre);

    /* Tell Bognor that we're now active */
    hrn_state_manager_set_can_show_visual (state_manager, TRUE);
  } else if (old_level == HRN_ZOOM_LEVEL_THEATRE &&
             new_level != HRN_ZOOM_LEVEL_THEATRE) {
    hrn_theatre_hide ((HrnTheatre *) hrn_theatre);
    hrn_ui_controls_show (controls);

    clutter_actor_show ((ClutterActor *)hrn_view);
    clutter_actor_animate (hrn_view, CLUTTER_LINEAR,
                           HRN_FROM_THEATRE_DURATION,
                           "opacity", 0xFF,
                           NULL);

    hrn_state_manager_set_can_show_visual (state_manager, FALSE);
  }

  switch (new_level)
    {
      case HRN_ZOOM_LEVEL_OVERVIEW:
        zl = 0.5;
        break;

      case HRN_ZOOM_LEVEL_COLLECTIONS:
        zl = 1.0;
        break;

      case HRN_ZOOM_LEVEL_SMALL_ITEMS:
        zl = 2.0;
        break;

      case HRN_ZOOM_LEVEL_LARGE_ITEMS:
        zl = 4.0;
        break;

      default:
        zl = 8.0;
        /* FIXME: This should be done by the theatre */
        hrn_controls_wake ();
        break;
    }

  hrn_set_zoom (zl);
  hrn_state_manager_set_zoom (state_manager, zl);

  /* FIXME: This should be done by the toolbar,
     checking if a current zoom level is pinned */
  /* hrn_toolbar_set_pinned (HRN_TOOLBAR (hrn_toolbar), FALSE); */
}

void
hrn_loc_slider_update ()
{
  hrn_ui_controls_set_position ((HrnUiControls *) hrn_ui_controls,
                                hrn_view_get_loc ());
}

/* FIXME: WTF is this for? */
gint desired_pos = 0;

static void
scrollbar_changed (HrnUiControls  *controls,
                   double          value,
                   gpointer        view)
{
  if (!(value >= 0.0 && value <= 1.0)) {
    g_warning ("value %f is out of range for scrollbar", value);
    value = 0.0;
  }

  hrn_view_set_loc (value);

  desired_pos = hrn_view_get_focused (HRN_VIEW (hrn_view));
}

void sidebar_init (HrnSidebar *sidebar); /*XXX: refactor into something sane */

static gboolean
theatre_key_press_cb (ClutterActor *actor, ClutterEvent *event,
                      gpointer userdata)
{
  switch (clutter_event_get_key_symbol ((ClutterEvent *) event))
    {
      case CLUTTER_space:
        hrn_queue_set_playing (master_queue,
                               !hrn_queue_get_playing (master_queue));
        break;

      default:
        break;
    }

  return FALSE;
}

static void
quit_requested_cb (HrnToolbar *toolbar,
                   gpointer    userdata)
{
  hrn_quit ();
}

/* FIXME: Temporary until we can move all the UI code into its own object */
static guint search_handler = 0;
static gboolean
deferred_search (gpointer data)
{
  HrnUiControls *controls = (HrnUiControls *) data;
  HrnToolbar *toolbar = hrn_ui_controls_get_toolbar (controls);
  const char *search;

  search_handler = 0;

  search = hrn_toolbar_get_query (toolbar);
  if (search && *search != '\0')
    {
      hrn_view_set_search (HRN_VIEW (hrn_view), search);
    }
  else
    {
      hrn_view_set_search (HRN_VIEW (hrn_view), NULL);
    }

  return FALSE;
}

static void
query_changed_cb (HrnUiControls *controls,
                  const char    *query,
                  gpointer       userdata)
{
  if (search_handler)
    {
      g_source_remove (search_handler);
      search_handler = 0;
    }

  hrn_state_manager_set_query (state_manager, query);
  if (query == NULL || *query == '\0')
    {
      deferred_search (controls);
    }

  search_handler = g_timeout_add (250, deferred_search, controls);
}

static void
source_changed_cb (HrnUiControls *controls,
                   HrnSource     *source,
                   gpointer       userdata)
{
  const char *source_path;

  hrn_view_set_source ((HrnView *) hrn_view, source);

  source_path = hrn_source_get_object_path (source);
  hrn_state_manager_set_source (state_manager, source_path);
}

static void
filter_changed_cb (HrnUiControls *controls,
                   int            filter,
                   gpointer       userdata)
{
  hrn_view_set_filter (HRN_VIEW (hrn_view), filter);
  hrn_view_center (hrn_view, 0);

  hrn_state_manager_set_filter (state_manager, filter);
}

static void
hrn_build_ui (void)
{
  ClutterActor *child = clutter_group_new ();
  int width, height;

  gtk_window_get_size (GTK_WINDOW (window), &width, &height);

  clutter_group_add (child, hrn_content_area = clutter_group_new ());
  clutter_actor_set_position (hrn_content_area, 180, 0);
  clutter_group_add (CLUTTER_GROUP (hrn_content_area),
                     hrn_view = hrn_view_new ());

  clutter_group_add (CLUTTER_GROUP (child), hrn_theatre_parent =
                     clutter_group_new ());

  hrn_theatre = CLUTTER_ACTOR (hrn_theatre_new (state_manager));
  clutter_group_add (CLUTTER_GROUP (hrn_theatre_parent), hrn_theatre);

  /* Connect to this here to trap the Space keypress because the theatre
     doesn't know (or care) whether its playing or not */
  g_signal_connect (hrn_theatre, "key-press-event",
                    G_CALLBACK (theatre_key_press_cb), NULL);

  hrn_ui_controls = g_object_new (HRN_TYPE_UI_CONTROLS,
                                  "pin-manager", pin_manager,
                                  "local-queue", local_queue,
                                  NULL);
  clutter_group_add (child, hrn_ui_controls);
  clutter_actor_set_size (hrn_ui_controls, width, height);

  g_signal_connect (hrn_ui_controls, "position-changed",
                    G_CALLBACK (scrollbar_changed), NULL);
  g_signal_connect (hrn_ui_controls, "zoom-changed",
                    G_CALLBACK (zoom_changed_cb), NULL);
  g_signal_connect (hrn_ui_controls, "quit-requested",
                    G_CALLBACK (quit_requested_cb), NULL);
  g_signal_connect (hrn_ui_controls, "filter-changed",
                    G_CALLBACK (filter_changed_cb), NULL);
  g_signal_connect (hrn_ui_controls, "source-changed",
                    G_CALLBACK (source_changed_cb), NULL);
  g_signal_connect (hrn_ui_controls, "query-changed",
                    G_CALLBACK (query_changed_cb), NULL);

  {
    HrnControls *controls = hrn_controls_new (local_queue);

    master_controls = CLUTTER_ACTOR (controls);
    clutter_group_add (CLUTTER_GROUP (hrn_theatre_parent), master_controls);
    clutter_actor_set_position (master_controls, 0, height - 60);

    clutter_actor_hide (hrn_theatre_parent);
    clutter_actor_hide (master_controls);
  }

  g_signal_connect (hrn_window_get_stage (window), "captured-event",
                    G_CALLBACK (controls_capture_cb), NULL);

  clutter_actor_hide (hrn_theatre);

  {
    ClutterColor shade = {0,0,0,0x88};
    hrn_preload_block = clutter_rectangle_new ();
    clutter_actor_set_size (hrn_preload_block, 4000, 4000);
    clutter_rectangle_set_color (CLUTTER_RECTANGLE (hrn_preload_block), &shade);
    clutter_actor_set_reactive (hrn_preload_block, TRUE);
    clutter_group_add (child, hrn_preload_block);
  }

  hrn_preload_content_block = clutter_rectangle_new ();
  clutter_actor_set_size (hrn_preload_content_block, 4000, 4000);
  clutter_actor_set_position (hrn_preload_content_block, -500, -500);
  clutter_group_add (hrn_content_area, hrn_preload_content_block);

  hrn_window_set_internal_actor (window, child);
}

#define CLUSTERS    10



static gboolean
key_pressed (ClutterActor *actor, ClutterEvent *event, gpointer data)
{
  switch (clutter_event_get_key_symbol ((ClutterEvent *) event))
    {
      case CLUTTER_Return:
        break;

      case CLUTTER_Page_Up:
        hrn_view_scroll (-1.0);
        break;

      case CLUTTER_Page_Down:
        hrn_view_scroll (1.0);
        break;

      case CLUTTER_Up:
        hrn_view_scroll (-0.5);
        break;

      case CLUTTER_Down:
        hrn_view_scroll (0.5);
        break;

      case CLUTTER_Escape:
        if (hrn_get_zoom () > 0.5)
           hrn_set_zoom (hrn_get_zoom () / 2.0);
        break;

      case CLUTTER_Right:
        /*target_scale *= 1.4142;*/
        if (hrn_theatre_get_active (HRN_THEATRE (hrn_theatre)))
          {
          }
        else
          {
            hrn_set_zoom (hrn_get_zoom () * 2.0);
          }
        break;

      case CLUTTER_Left:
        if (hrn_theatre_get_active (HRN_THEATRE (hrn_theatre)))
          {
            /* this goes to the context/thumbbar item left of
             * the current one
             */
          }
        else
          {
            if (hrn_get_zoom () > 0.5)
              hrn_set_zoom (hrn_get_zoom () / 2.0);
          }
        break;

      default:
        break;
    }
  return FALSE;
}

static gchar *load_path = NULL;

static gboolean
auto_zoomer (gpointer userdata)
{
  static gfloat target_zoom = 0.0;

  if (hrn_get_scale () != target_zoom && target_zoom > 0.0)
    {
    }
  else
    {
      if (target_zoom == 8.0)
        {
          target_zoom = 0.5;
        }
      else
        {
          target_zoom = 8.0;
        }
      hrn_texture_cache_evict_items (hrn_texture_cache_get_default (), 1000);
      hrn_set_zoom (target_zoom);
    }
  return TRUE;
}

gboolean hrn_in_preload = FALSE;
static gboolean preload_done = FALSE;

static gboolean
hrn_preload_data (gpointer userdata)
{
  static gint iteration = 0;

  iteration ++;

  /* Do test renders at specific zoom levels, this forces
   * hornsey to have all the needed data for the transitions
   * cached.
   */
  switch (iteration)
    {
      case 1: hrn_set_zoom_hard (0.75); break;
      case 2: hrn_set_zoom_hard (1.25); break;
      case 3: hrn_set_zoom_hard (1.88); break;
      case 4: hrn_set_zoom_hard (2.0); break;
      case 5: hrn_set_zoom_hard (2.1); break;
      case 6: hrn_view_set_loc (0.4); hrn_set_zoom_hard (1.25); break;
      case 7: hrn_set_zoom_hard (1.88); break;
      case 8: hrn_set_zoom_hard (2.0); break;
      case 9: hrn_set_zoom_hard (1.0); break;
      default:
         hrn_view_set_loc (0.0);
	 hrn_set_zoom_hard (1.0);   /* this is the default starting zoom level */
	 preload_done = TRUE;
	 return FALSE;
    }
  clutter_actor_paint (hrn_view); /* this is needed to force a full
                                     pregeneration cycle */
  return TRUE;
}

static void
check_all_sources_ready (void)
{
  if (source_replies == source_count &&
      local_source != NULL)
    {
      const char *source_path;

      hrn_state_manager_ready (state_manager);

      if (hrn_state_manager_get_force_visual_mode (state_manager)) {
         hrn_set_theatre ();
      }

      hrn_view_start ();

      /* maybe the following two calls should be removed */
      source_path = hrn_state_manager_get_source (state_manager);
      if (source_path == NULL || *source_path == '\0' ||
          g_str_equal (source_path, BKL_LOCAL_SOURCE_PATH)) {
        hrn_view_set_source (HRN_VIEW (hrn_view), local_source);
      } else {
        HrnSource *source = hrn_get_source_for_path (source_path);

        if (source == NULL) {
          source = local_source;
        }

        hrn_view_set_source (HRN_VIEW (hrn_view), source);
      }

      if (load_path)
        {
          BklItem *item;

          g_debug ("Hornsey commandline URI: %s", load_path);

          item = hrn_get_item_for_uri (load_path);
          hrn_set_theatre ();
          if (item)
            {
              hrn_play_now (item);
            }
          else
            {
              hrn_play_uri_now (load_path);
            }
        }
      else if (!hrn_state_manager_get_force_visual_mode (state_manager)) {
        HrnToolbar *tb;

        /* If were not launching into playback mode assume
           browsing and need for preloading */

        hrn_in_preload = TRUE;

        if (hrn_get_zoom () < 8.0)
          {
            g_idle_add (hrn_preload_data, NULL);
            while (preload_done == FALSE) {
              g_main_context_iteration (NULL, FALSE);
            }
          }
        hrn_in_preload = FALSE;

        /* Restore state */
        hrn_set_zoom (hrn_state_manager_get_zoom (state_manager));
        hrn_view_set_filter ((HrnView *) hrn_view,
                             hrn_state_manager_get_filter (state_manager));

        tb = hrn_ui_controls_get_toolbar ((HrnUiControls *) hrn_ui_controls);
        hrn_toolbar_set_query (tb, hrn_state_manager_get_query (state_manager));
        hrn_toolbar_set_filter_mode
          (tb, hrn_state_manager_get_filter (state_manager));
      }

#ifdef HRN_SN_SUPPORT
      {
        SnLauncheeContext *context;

        context = g_object_get_data (G_OBJECT (window), "sn-context");
        if (context)
          {
            sn_launchee_context_complete (context);
            sn_launchee_context_unref (context);
          }
      }
#endif

      hrn_queue_populate (master_queue);
      hrn_controls_populate (HRN_CONTROLS (master_controls));

      clutter_actor_destroy (hrn_preload_content_block);
      clutter_actor_destroy (hrn_preload_block);

      if (hrn_auto_zoom)
        g_timeout_add (50, auto_zoomer, NULL);

#if 0
      if (load_path == NULL) {
          /* The state needs to be restored in an idle because of a bug in NBTK
             that the handle position of the zoombar gets set before the zoombar
             has an allocation, so the position is always set to 0.0 no matter
             what the actual Adjustment says:

             http://bugzilla.moblin.org/show_bug.cgi?id=3691 */
          g_idle_add (restore_state_idle, NULL);
      }
#endif
    }
}

static void
source_ready_cb (BklSourceClient *source_client, gpointer userdata)
{
  HrnSource *source;

  source_replies++;

  source = hrn_source_new (source_client);
  if (source)
    {
      HrnSidebar *sidebar = hrn_ui_controls_get_sidebar ((HrnUiControls *) hrn_ui_controls);

      transient_sources = g_list_prepend (transient_sources, source);
      hrn_sidebar_add_source (sidebar, source, FALSE);
    }
  else
    {
      g_warning ("Could not create source for '%s'",
                 bkl_source_client_get_path (source_client));
    }

  /* HrnSource takes its own reference, we want to unref whether
     hrn_source_new() succeeded or not */
  g_object_unref (source_client);

  check_all_sources_ready ();
}

static void
source_added (BklSourceManagerClient *client,
              const char             *object_path,
              gpointer                userdata)
{
  BklSourceClient *source;

  source = bkl_source_client_new (object_path);
  g_signal_connect (source, "ready", G_CALLBACK (source_ready_cb), NULL);
}

static void
source_removed (BklSourceManagerClient *client,
                const char             *object_path,
                gpointer                userdata)
{
  GList *s;

  for (s = transient_sources; s; s = s->next)
    {
      HrnSource *source = s->data;

      if (g_str_equal (object_path,
                       hrn_source_get_object_path (source)))
        {
          HrnSidebar *sidebar;

          if (hrn_view_get_source (HRN_VIEW (hrn_view)) == source)
            {
              HrnToolbar *tb;

              tb = hrn_ui_controls_get_toolbar ((HrnUiControls *) hrn_ui_controls);
              hrn_toolbar_set_pinned (tb, FALSE);

              hrn_view_set_source (HRN_VIEW (hrn_view), hrn_get_local_source ());
              hrn_view_set_search (HRN_VIEW (hrn_view), "");
              hrn_set_zoom_hard (0.5);
              hrn_view_set_filter (HRN_VIEW (hrn_view), BKL_ITEM_TYPE_IMAGE|
                                                        BKL_ITEM_TYPE_AUDIO|
                                                        BKL_ITEM_TYPE_VIDEO);
            }

          sidebar = hrn_ui_controls_get_sidebar ((HrnUiControls *) hrn_ui_controls);
          hrn_sidebar_remove_source (sidebar, source);

          g_object_unref (source);
          transient_sources = g_list_delete_link (transient_sources, s);
          break;
        }
    }
}

HrnSource *
hrn_get_local_source (void)
{
  return local_source;
}

GList *
hrn_get_transient_sources (void)
{
  return transient_sources;
}

HrnSource *
hrn_get_source_for_path (const char *path)
{
  GList *s;

  for (s = transient_sources; s; s = s->next)
    {
      HrnSource *source = s->data;

      if (g_str_equal (path, hrn_source_get_object_path (source)))
        {
          return source;
        }
    }

  return NULL;
}

BklItem *
hrn_get_item_for_uri (const char *uri)
{
  BklItem *item = NULL;
  GList   *s;

  /* Check local source first */
  item = hrn_source_get_item (local_source, uri);
  if (item)
    {
      return item;
    }

  /* Check transients */
  for (s = transient_sources; s; s = s->next)
    {
      HrnSource *source = s->data;

      item = hrn_source_get_item (source, uri);
      if (item)
        {
          return item;
        }
    }

  return NULL;
}

static void
local_source_ready (BklSourceClient *client)
{
  local_source = hrn_source_new (client);
  if (local_source)
    {
      HrnSidebar *sidebar = hrn_ui_controls_get_sidebar ((HrnUiControls *) hrn_ui_controls);
      hrn_sidebar_add_source (sidebar, local_source, TRUE);
    }
  else
    {
      g_warning ("Could not create a source for '%s'",
                 bkl_source_client_get_path (client));
    }

  /* HrnSource takes its own reference, we want to unref whether
     hrn_source_new() succeeded or not */
  g_object_unref (client);

  check_all_sources_ready ();
}

static void
get_sources_reply (BklSourceManagerClient *client, GList *sources,
                   GError                 *error,
                   gpointer userdata)
{
  BklSourceClient *ls;
  GList           *s;

  if (error != NULL)
    {
      g_warning ("Error getting sources: %s", error->message);
      g_error_free (error);
    }

  ls = bkl_source_client_new (BKL_LOCAL_SOURCE_PATH);
  g_signal_connect (ls, "ready", G_CALLBACK (local_source_ready), ls);

  for (s = sources; s; s = s->next)
    {
      source_count++;
      source_added (manager, s->data, NULL);
    }

  g_list_free (sources);
}

static void
setup_sources (BklSourceManagerClient *manager_client, gpointer userdata)
{
  bkl_source_manager_client_get_sources (manager, get_sources_reply, NULL);
}

void
hrn_play_uri_now (const gchar *uri)
{
  hrn_queue_play_uri_now (master_queue, uri);
}

void
hrn_play_now (BklItem *item)
{
  hrn_queue_play_now (master_queue, item);
}

void
hrn_stage_remove_decorations (ClutterStage *stage)
{
  struct
  {
    unsigned long flags;
    unsigned long functions;
    unsigned long decorations;
    long          inputMode;
    unsigned long status;
  }        MWMHints = { 2, 0, 0, 0, 0 };

  Display *xdisplay = clutter_x11_get_default_display ();
  Atom     wm_hints = XInternAtom (xdisplay, "_MOTIF_WM_HINTS", True);
  Window   xwindow  = clutter_x11_get_stage_window (CLUTTER_STAGE (stage));

  if (wm_hints != None)
    {
      XChangeProperty (xdisplay, xwindow, wm_hints, wm_hints, 32,
                       PropModeReplace, (guchar*) &MWMHints,
                       sizeof (MWMHints) / sizeof (long));
    }
}

static void
dbus_init (HrnStateManager *sm)
{
  DBusGConnection *connection;
  GError          *error = NULL;

  connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (connection == NULL)
    {
      g_warning ("Error getting bus: %s", error->message);
      g_error_free (error);
      return;
    }

  /* We dont need to request the horney name as libunique has already
     done that for us */

  dbus_g_connection_register_g_object (connection, HORNSEY_DBUS_PATH,
                                       (GObject *) sm);
}

static GdkFilterReturn
gdk_to_clutter_event_forward (GdkXEvent *xevent,
                              GdkEvent  *event,
                              gpointer   data)
{
  switch (clutter_x11_handle_event ((XEvent *) xevent)) {
  default:
  case CLUTTER_X11_FILTER_CONTINUE:
    return GDK_FILTER_CONTINUE;
  case CLUTTER_X11_FILTER_TRANSLATE:
    return GDK_FILTER_TRANSLATE;
  case CLUTTER_X11_FILTER_REMOVE:
    return GDK_FILTER_REMOVE;
  }

  return GDK_FILTER_CONTINUE;
}

gint            DIM;
gboolean        hrn_auto_zoom = FALSE;

gint
main (gint argc, gchar **argv)
{
  UniqueApp   *app;

  gfloat       stage_width      = 0;
  gfloat       stage_height     = 0;
#ifdef HAVE_WINDOWMODE
  gboolean     stage_fullscreen = FALSE;
#else
  gboolean     stage_fullscreen = TRUE;
#endif

  GError *error = NULL;
  static gboolean window_mode = FALSE;
  static char *geometry = NULL;
  static char **files = NULL;

#ifdef HRN_SN_SUPPORT
  SnLauncheeContext *context;
  SnDisplay *display;
  Display *xdisplay;
  int screen;
#endif

  static GOptionEntry hrn_options[] = {
    { "window", 'w', 0, G_OPTION_ARG_NONE, &window_mode,
      N_("Run in a window"), NULL },
    { "size", 's', 0, G_OPTION_ARG_STRING, &geometry,
      N_("Size of the window"), "widthXheight" },
    { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &files,
      NULL, N_("[FILE...]") },
    { NULL },
  };

  bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);

  g_thread_init (NULL);

  if (gtk_clutter_init_with_args (&argc, &argv, "Hornsey media player",
                                  hrn_options, NULL, &error) <= 0)
  {
      g_printf ("Error: %s\n", error->message);
      g_error_free (error);
      return 1;
  }
  gdk_window_add_filter (NULL, gdk_to_clutter_event_forward, NULL);

  clutter_gst_init (&argc, &argv);

  bkl_init ();
  notify_init ("Hornsey Media Player");

  clutter_set_font_flags (CLUTTER_FONT_HINTING);

  g_set_application_name ("Hornsey Media Player");

#ifdef HRN_SN_SUPPORT
  xdisplay = clutter_x11_get_default_display ();

  if (g_getenv ("LIBSN_SYNC") != NULL)
    XSynchronize (xdisplay, True);

  display = sn_display_new (xdisplay, NULL, NULL);
  screen  = clutter_x11_get_default_screen ();
  context = sn_launchee_context_new_from_environment (display, screen);
#endif

  app = unique_app_new ("org.moblin.Hornsey", NULL);

  if (unique_app_is_running (app))
    {
      if (files == NULL)
        {
          unique_app_send_message (app, UNIQUE_ACTIVATE, NULL);
        }
      else
        {
          UniqueMessageData *data = unique_message_data_new ();

          /* FIXME: Send all the uris */
          unique_message_data_set_text (data, files[0], -1);
          unique_app_send_message (app, UNIQUE_OPEN, data);
          unique_message_data_free (data);
        }
      return 0;
    }

  state_manager = g_object_new (HRN_TYPE_STATE_MANAGER, NULL);
  dbus_init (state_manager);

  local_queue = br_queue_new_local ();

  nbtk_style_load_from_file (nbtk_style_get_default (),
                             PKGDATADIR "hornsey.css", NULL);

  if (window_mode == TRUE) {
    stage_fullscreen = FALSE;
  }

  if (geometry != NULL) {
    char *xspot;

    stage_width = atoi (geometry);

    xspot = strchr (geometry, 'x');
    if (xspot == NULL) {
      xspot = strchr (geometry, 'X');
    }

    if (xspot) {
      stage_height = atoi (xspot + 1);
    }
  }

  if (files != NULL) {
    /* FIXME: Handle all the uris */
    load_path = g_strdup (files[0]);
  }

  window = hrn_window_new (state_manager);
  g_signal_connect (window, "destroy",
                    G_CALLBACK (gtk_main_quit), NULL);

  if (!stage_fullscreen)
    {
      stage_width  = 800;
      stage_height = 480;
    }
  else if (stage_width == 0 &&
           stage_height == 0)
    {
      /* This causes us to read back the dimensions of the X desktop
       * using only Clutter APIs
       */
      stage_width  = gdk_screen_get_width (gdk_screen_get_default ());
      stage_height = gdk_screen_get_height (gdk_screen_get_default ());
    }

  gtk_window_resize (GTK_WINDOW (window), stage_width, stage_height);
  if (stage_fullscreen) {
    hrn_window_fullscreen (HRN_WINDOW (window));
  }

#ifdef HRN_SN_SUPPORT
  if (context) {
    g_object_set_data (G_OBJECT (window), "sn-context", context);
  }
#endif

  DIM = (stage_width - 190) / 4.25;

  {
    gchar *search_path = hrn_make_config_file ("queries");

    /* Check for the old skool version as well */
    if (g_file_test (search_path, G_FILE_TEST_EXISTS) == FALSE) {
      g_free (search_path);

      g_warning ("Loading old deprecated query file");
      search_path = g_strdup_printf ("%s/.hrn-queries", g_get_home_dir ());
    }

    pin_manager = hrn_pin_manager_new (search_path);
    g_free (search_path);
  }

  hrn_build_ui ();

  /* Setup the Bickley Source manager */
  manager = g_object_new (BKL_TYPE_SOURCE_MANAGER_CLIENT, NULL);
  g_signal_connect (manager, "ready", G_CALLBACK (setup_sources), NULL);

  g_signal_connect (manager, "source-added",
                    G_CALLBACK (source_added), NULL);
  g_signal_connect (manager, "source-removed",
                    G_CALLBACK (source_removed), NULL);

#if HRN_DEBUG_OFF_STAGE
  {
    ClutterColor  fake_stage_color   = { 0x00, 0x00, 0x00, 0x00 };
    ClutterColor  stage_border_color = { 0xff, 0x66, 0x22, 0x22 };
    ClutterActor *fake_stage;
    fake_stage = clutter_rectangle_new ();
    clutter_actor_set_size (fake_stage, stage_width, stage_height);
    clutter_rectangle_set_border_color (CLUTTER_RECTANGLE (
                                          fake_stage), &stage_border_color);
    clutter_rectangle_set_border_width (CLUTTER_RECTANGLE (fake_stage), 20);
    clutter_rectangle_set_color (CLUTTER_RECTANGLE (
                                   fake_stage), &fake_stage_color);
    clutter_group_add (CLUTTER_GROUP (hrn_content_area), fake_stage);
    clutter_actor_set_depth (hrn_view, HRN_DEBUG_DEPTH);
    clutter_actor_set_depth (fake_stage, HRN_DEBUG_DEPTH);
  }
#endif

  g_signal_connect (hrn_window_get_stage (window), "key-press-event",
                    G_CALLBACK (key_pressed), hrn_view);

  gtk_window_set_title (GTK_WINDOW (window), _ ("Media Player"));

  g_signal_connect (app, "message-received",
                    G_CALLBACK (hrn_message_received_cb), NULL);

  gtk_widget_show_all (GTK_WIDGET (window));
  gtk_main ();

  /* Tell Bognor that we can't show visuals anymore */
  br_iface_player_emit_can_show_visual_changed (BR_IFACE_PLAYER (state_manager),
                                                FALSE);

  {
    char *query_path = hrn_make_config_file ("queries");

    hrn_pin_manager_write_to_file (pin_manager, query_path);
    g_free (query_path);
  }

  {
    char *state_path = hrn_make_config_file ("saved-state");

    hrn_state_manager_write_to_file (state_manager, state_path);
    g_free (state_path);
  }

  bkl_shutdown ();

  return 0;
}

char *
hrn_make_config_file (const char *filename)
{
  const char *base;
  char *path, *full;

  base = g_getenv ("XDG_CONFIG_HOME");
  if (base) {
    path = g_strdup_printf ("%s/hornsey", base);
    full = g_strdup_printf ("%s/hornsey/%s", base, filename);
  } else {
    path = g_strdup_printf ("%s/.config/hornsey", g_get_home_dir ());
    full = g_strdup_printf ("%s/.config/hornsey/%s", g_get_home_dir (),
                            filename);
  }

  /* Make sure the directory exists */
  if (g_file_test (path, G_FILE_TEST_EXISTS) == FALSE) {
    g_mkdir_with_parents (path, 0700);
  }
  g_free (path);

  return full;
}

void
hrn_quit (void)
{
  gtk_main_quit ();
}

void
hrn_set_zoom (double zoom)
{
  HrnZoomLevel zoom_level;

  if (hrn_in_preload == FALSE) {
    if (ABS (zoom - 0.5) < 0.01) {
      zoom_level = HRN_ZOOM_LEVEL_OVERVIEW;
    } else if (ABS (zoom - 1.0) < 0.01) {
      zoom_level = HRN_ZOOM_LEVEL_COLLECTIONS;
    } else if (ABS (zoom - 2.0) < 0.01) {
      zoom_level = HRN_ZOOM_LEVEL_SMALL_ITEMS;
    } else if (ABS (zoom - 4.0) < 0.01) {
      zoom_level = HRN_ZOOM_LEVEL_LARGE_ITEMS;
    } else {
      zoom_level = HRN_ZOOM_LEVEL_THEATRE;
    }

    hrn_ui_controls_set_zoom ((HrnUiControls *) hrn_ui_controls, zoom_level);
  }

  hrn_view_set_zoom ((HrnView *) hrn_view, zoom);
}

void
hrn_set_query (const char *query)
{
  hrn_ui_controls_set_query ((HrnUiControls *) hrn_ui_controls, query);
}
