#include <bognor/br-queue.h>
#include <nbtk/nbtk.h>

#include "hrn.h"
#include "hrn-toolbar.h"
#include "hrn-sidebar.h"
#include "hrn-pin-manager.h"
#include "hrn-ui-controls.h"
#include "hrn-queue.h"
#include "hrn-marshal.h"

enum {
    PROP_0,
    PROP_PIN_MANAGER,
    PROP_QUEUE,
};

enum {
    QUIT_REQUEST,
    POS_CHANGED,
    ZOOM_CHANGED,
    FILTER_CHANGED,
    SOURCE_CHANGED,
    QUERY_CHANGED,
    LAST_SIGNAL
};

struct _HrnUiControlsPrivate {
    HrnPinManager *pin_manager;

    NbtkWidget *toolbar;
    HrnZoomLevel current_zoom;

    HrnSidebar *sidebar;

    NbtkAdjustment *adjustment;
    guint32 adj_changed_id;

    NbtkWidget *scrollbar;

    ClutterActor *lowlight;

    BrQueue *local_queue;
    ClutterActor *queues;
    HrnQueue *queue;
};

#define GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), HRN_TYPE_UI_CONTROLS, HrnUiControlsPrivate))
G_DEFINE_TYPE (HrnUiControls, hrn_ui_controls, CLUTTER_TYPE_GROUP);

static guint32 signals[LAST_SIGNAL] = {0, };

static void
hrn_ui_controls_finalize (GObject *object)
{
    G_OBJECT_CLASS (hrn_ui_controls_parent_class)->finalize (object);
}

static void
hrn_ui_controls_dispose (GObject *object)
{
    HrnUiControls *self = (HrnUiControls *) object;
    HrnUiControlsPrivate *priv = self->priv;

    if (priv->pin_manager) {
        g_object_unref (priv->pin_manager);
        priv->pin_manager = NULL;
    }

    if (priv->local_queue) {
        g_object_unref (priv->local_queue);
        priv->local_queue = NULL;
    }

    G_OBJECT_CLASS (hrn_ui_controls_parent_class)->dispose (object);
}

static void
hrn_ui_controls_set_property (GObject      *object,
                              guint         prop_id,
                              const GValue *value,
                              GParamSpec   *pspec)
{
    HrnUiControls *self = (HrnUiControls *) object;
    HrnUiControlsPrivate *priv = self->priv;

    switch (prop_id) {
    case PROP_PIN_MANAGER:
        priv->pin_manager = g_value_dup_object (value);
        break;

    case PROP_QUEUE:
        priv->local_queue = g_value_dup_object (value);
        break;

    default:
        break;
    }
}

static void
hrn_ui_controls_get_property (GObject    *object,
                              guint       prop_id,
                              GValue     *value,
                              GParamSpec *pspec)
{
    switch (prop_id) {

    default:
        break;
    }
}

static void
quit_requested_cb (HrnToolbar    *toolbar,
                   HrnUiControls *uic)
{
    g_signal_emit (uic, signals[QUIT_REQUEST], 0);
}

static void
filter_changed_cb (HrnToolbar    *toolbar,
                   int            filter,
                   HrnUiControls *controls)
{
    g_signal_emit (controls, signals[FILTER_CHANGED], 0, filter);
}

static void
query_changed_cb (HrnToolbar    *toolbar,
                  const char    *query,
                  HrnUiControls *controls)
{
    g_signal_emit (controls, signals[QUERY_CHANGED], 0, query);
}

static void
loc_changed_cb (NbtkAdjustment *adjustment,
                GParamSpec     *pspec,
                HrnUiControls  *controls)
{
    double value = nbtk_adjustment_get_value (adjustment);

    g_signal_emit (controls, signals[POS_CHANGED], 0, value);
}

static void
zoom_changed_cb (HrnToolbar    *toolbar,
                 double         zoom,
                 HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;
    HrnZoomLevel old_zoom = priv->current_zoom;
    HrnZoomLevel new_zoom;

    new_zoom = (HrnZoomLevel) (zoom * 4.0);
    if (old_zoom == new_zoom) {
        return;
    }

    priv->current_zoom = new_zoom;
    g_signal_emit (controls, signals [ZOOM_CHANGED], 0, old_zoom, new_zoom);
}

static void
set_filter (HrnUiControls *controls,
            int            filter)
{
    HrnUiControlsPrivate *priv = controls->priv;

    hrn_toolbar_set_filter_mode ((HrnToolbar *) priv->toolbar, filter);
    g_signal_emit (controls, signals[FILTER_CHANGED], 0, filter);
}

static void
set_source (HrnUiControls *controls,
            HrnSource     *source)
{
    g_signal_emit (controls, signals[SOURCE_CHANGED], 0, source);
}

static void
sidebar_source_changed_cb (HrnSidebar    *sidebar,
                           HrnSource     *source,
                           int            filter,
                           HrnPin        *pin,
                           HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;

    set_filter (controls, filter);
    set_source (controls, source);

    if (pin == NULL) {
        hrn_toolbar_set_query ((HrnToolbar *) priv->toolbar, "");
        hrn_pin_manager_select_pin (priv->pin_manager, NULL);
        hrn_set_zoom_hard (0.5);
        return;
    }

    hrn_toolbar_set_query ((HrnToolbar *) priv->toolbar, pin->query);
    hrn_pin_manager_select_pin (priv->pin_manager, pin);
    hrn_set_zoom_hard ((float) pin->zoom);
}

static GObject *
hrn_ui_controls_constructor (GType                  type,
                             guint                  n_params,
                             GObjectConstructParam *params)
{
    GObject *object;
    HrnUiControls *controls;
    HrnUiControlsPrivate *priv;
    ClutterColor  black = { 0x00, 0x00, 0x00, 0xff };

    object = G_OBJECT_CLASS (hrn_ui_controls_parent_class)->constructor
        (type, n_params, params);

    controls = HRN_UI_CONTROLS (object);
    priv = controls->priv;

    priv->toolbar = hrn_toolbar_new (priv->pin_manager);
    clutter_container_add_actor ((ClutterContainer *) controls,
                                 (ClutterActor *) priv->toolbar);
    priv->current_zoom = hrn_toolbar_get_zoom ((HrnToolbar *) priv->toolbar);
    g_signal_connect (priv->toolbar, "query-changed",
                      G_CALLBACK (query_changed_cb), controls);
    g_signal_connect (priv->toolbar, "quit-requested",
                      G_CALLBACK (quit_requested_cb), controls);
    g_signal_connect (priv->toolbar, "filter-changed",
                      G_CALLBACK (filter_changed_cb), controls);
    g_signal_connect (priv->toolbar, "zoom-changed",
                      G_CALLBACK (zoom_changed_cb), controls);

    priv->sidebar = hrn_sidebar_new (priv->pin_manager);
    clutter_container_add_actor ((ClutterContainer *) controls,
                                 (ClutterActor *) priv->sidebar);
    g_signal_connect (priv->sidebar, "source-changed",
                      G_CALLBACK (sidebar_source_changed_cb), controls);

    priv->adjustment = nbtk_adjustment_new (0.0, 0.0, 1.0, 0.2, 0.0, 0.0);
    priv->scrollbar = nbtk_scroll_bar_new (NULL);
    clutter_container_add_actor ((ClutterContainer *) controls,
                                 (ClutterActor *) priv->scrollbar);
    g_object_set (priv->scrollbar,
                  "vertical", TRUE,
                  NULL);
    nbtk_scroll_bar_set_adjustment (NBTK_SCROLL_BAR (priv->scrollbar),
                                    priv->adjustment);
    priv->adj_changed_id = g_signal_connect (priv->adjustment, "notify::value",
                                             G_CALLBACK (loc_changed_cb),
                                             controls);

    priv->lowlight = clutter_rectangle_new_with_color (&black);
    clutter_container_add_actor ((ClutterContainer *) controls, priv->lowlight);
    clutter_actor_hide (priv->lowlight);
    clutter_actor_set_opacity (priv->lowlight, 0x00);

    /* Should be a custom actor */
    priv->queues = clutter_group_new ();
    clutter_container_add_actor ((ClutterContainer *) controls, priv->queues);

    master_queue = priv->queue = hrn_queue_new (priv->local_queue);
    clutter_container_add_actor ((ClutterContainer *) priv->queues,
                                 (ClutterActor *) priv->queue);

    return object;
}

#define TOOLBAR_HEIGHT 62.0
#define SCROLLBAR_WIDTH 22.0

static void
hrn_ui_controls_allocate (ClutterActor           *self,
                          const ClutterActorBox  *box,
                          ClutterAllocationFlags  flags)
{
    HrnUiControls *controls = (HrnUiControls *) self;
    HrnUiControlsPrivate *priv = controls->priv;
    float width, height;

    width = box->x2 - box->x1;
    height = box->y2 - box->y1;

    clutter_actor_set_size (CLUTTER_ACTOR (priv->scrollbar),
                            SCROLLBAR_WIDTH, height - TOOLBAR_HEIGHT);
    clutter_actor_set_position (CLUTTER_ACTOR (priv->scrollbar),
                                width - SCROLLBAR_WIDTH, TOOLBAR_HEIGHT);

    clutter_actor_set_y (CLUTTER_ACTOR (priv->sidebar), TOOLBAR_HEIGHT);
    clutter_actor_set_position (CLUTTER_ACTOR (priv->queues),
                                width - 335, height - 60);

    clutter_actor_set_size (priv->lowlight, width, height);

    CLUTTER_ACTOR_CLASS (hrn_ui_controls_parent_class)->allocate
        (self, box, flags);
}

static void
hrn_ui_controls_class_init (HrnUiControlsClass *klass)
{
    GObjectClass *o_class = (GObjectClass *) klass;
    ClutterActorClass *a_class = (ClutterActorClass *) klass;

    o_class->dispose = hrn_ui_controls_dispose;
    o_class->finalize = hrn_ui_controls_finalize;
    o_class->set_property = hrn_ui_controls_set_property;
    o_class->get_property = hrn_ui_controls_get_property;
    o_class->constructor = hrn_ui_controls_constructor;

    a_class->allocate = hrn_ui_controls_allocate;

    g_type_class_add_private (klass, sizeof (HrnUiControlsPrivate));

    g_object_class_install_property (o_class, PROP_PIN_MANAGER,
                                     g_param_spec_object ("pin-manager", "", "",
                                                          HRN_TYPE_PIN_MANAGER,
                                                          G_PARAM_WRITABLE |
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_STATIC_STRINGS));
    g_object_class_install_property (o_class, PROP_QUEUE,
                                     g_param_spec_object ("local-queue", "", "",
                                                          BR_TYPE_QUEUE,
                                                          G_PARAM_WRITABLE |
                                                          G_PARAM_CONSTRUCT_ONLY |
                                                          G_PARAM_STATIC_STRINGS));

    signals[QUIT_REQUEST] = g_signal_new ("quit-requested",
                                          G_TYPE_FROM_CLASS (klass),
                                          G_SIGNAL_RUN_FIRST |
                                          G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                          g_cclosure_marshal_VOID__VOID,
                                          G_TYPE_NONE, 0);
    signals[POS_CHANGED] = g_signal_new ("position-changed",
                                         G_TYPE_FROM_CLASS (klass),
                                         G_SIGNAL_RUN_FIRST |
                                         G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                         g_cclosure_marshal_VOID__DOUBLE,
                                         G_TYPE_NONE, 1, G_TYPE_DOUBLE);
    signals[ZOOM_CHANGED] = g_signal_new ("zoom-changed",
                                          G_TYPE_FROM_CLASS (klass),
                                          G_SIGNAL_RUN_FIRST |
                                          G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                          hrn_marshal_VOID__INT_INT,
                                          G_TYPE_NONE, 2, G_TYPE_INT,
                                          G_TYPE_INT);
    signals[SOURCE_CHANGED] = g_signal_new ("source-changed",
                                            G_TYPE_FROM_CLASS (klass),
                                            G_SIGNAL_RUN_FIRST |
                                            G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                            g_cclosure_marshal_VOID__POINTER,
                                            G_TYPE_NONE, 1, G_TYPE_POINTER);
    signals[FILTER_CHANGED] = g_signal_new ("filter-changed",
                                            G_TYPE_FROM_CLASS (klass),
                                            G_SIGNAL_RUN_FIRST |
                                            G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                            g_cclosure_marshal_VOID__INT,
                                            G_TYPE_NONE, 1, G_TYPE_INT);
    signals[QUERY_CHANGED] = g_signal_new ("query-changed",
                                           G_TYPE_FROM_CLASS (klass),
                                           G_SIGNAL_RUN_FIRST |
                                           G_SIGNAL_NO_RECURSE, 0, NULL, NULL,
                                           g_cclosure_marshal_VOID__STRING,
                                           G_TYPE_NONE, 1, G_TYPE_STRING);
}

static void
hrn_ui_controls_init (HrnUiControls *self)
{
    HrnUiControlsPrivate *priv;

    self->priv = GET_PRIVATE (self);
    priv = self->priv;
}

void
hrn_ui_controls_set_position (HrnUiControls *controls,
                              double         value)
{
    HrnUiControlsPrivate *priv = controls->priv;

    /* FIXME: In an ideal world, this would be blocked here
       but the view requires this to be emitted so that it scrolls correctly */
    /* g_signal_handler_block (priv->adjustment, priv->adj_changed_id); */
    nbtk_adjustment_set_value (priv->adjustment, value);
    /* g_signal_handler_unblock (priv->adjustment, priv->adj_changed_id); */
}

HrnToolbar *
hrn_ui_controls_get_toolbar (HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;

    return (HrnToolbar *) priv->toolbar;
}

HrnSidebar *
hrn_ui_controls_get_sidebar (HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;

    return priv->sidebar;
}

HrnZoomLevel
hrn_ui_controls_get_zoom (HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;

    return priv->current_zoom;
}

void
hrn_ui_controls_set_zoom (HrnUiControls *controls,
                          HrnZoomLevel   zoom_level)
{
    HrnUiControlsPrivate *priv = controls->priv;

    if (zoom_level == priv->current_zoom) {
        return;
    }

    hrn_toolbar_set_zoom ((HrnToolbar *) priv->toolbar,
                          ((double) zoom_level) / 4.0);
}

static void
hide_actor (ClutterAnimation *anim,
            ClutterActor     *actor)
{
    clutter_actor_hide (actor);
}

void
hrn_ui_controls_hide (HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;
    float width;

    clutter_actor_animate ((ClutterActor *) controls, CLUTTER_LINEAR,
                           HRN_TO_THEATRE_DURATION,
                           "opacity", 0x00,
                           "signal::completed", hide_actor, controls,
                           NULL);

    width = clutter_actor_get_width ((ClutterActor *) controls);

    clutter_actor_animate ((ClutterActor *) priv->scrollbar,
                           CLUTTER_EASE_IN_OUT_CUBIC,
                           HRN_TO_THEATRE_DURATION,
                           "x", width,
                           NULL);
}

void
hrn_ui_controls_show (HrnUiControls *controls)
{
    HrnUiControlsPrivate *priv = controls->priv;
    float width;

    /* going from theatre mode */

    clutter_actor_show ((ClutterActor *) controls);

    width = clutter_actor_get_width ((ClutterActor *) controls);
    clutter_actor_animate ((ClutterActor *) priv->scrollbar,
                           CLUTTER_EASE_IN_OUT_CUBIC,
                           HRN_FROM_THEATRE_DURATION,
                           "x", width,
                           NULL);
    clutter_actor_animate ((ClutterActor *) controls, CLUTTER_LINEAR,
                           HRN_FROM_THEATRE_DURATION,
                           "opacity", 0xFF,
                           NULL);

    hrn_toolbar_focused ((HrnToolbar *) priv->toolbar);
}

void
hrn_ui_controls_set_query (HrnUiControls *controls,
                           const char    *query)
{
    HrnUiControlsPrivate *priv = controls->priv;

    hrn_toolbar_set_query ((HrnToolbar *) priv->toolbar, query);
}

static void
hide_when_done (ClutterAnimation *animation,
                ClutterActor     *actor)
{
  clutter_actor_hide (actor);
}

void
hrn_ui_controls_set_enabled (HrnUiControls *controls,
                             gboolean       enabled)
{
    HrnUiControlsPrivate *priv = controls->priv;

    if (enabled == FALSE) {
        clutter_actor_show (priv->lowlight);
        clutter_actor_animate (priv->lowlight, CLUTTER_LINEAR,
                               500, "opacity", 0x5c,
                               NULL);
    } else {
        clutter_actor_animate (priv->lowlight, CLUTTER_LINEAR, 500,
                               "opacity", 0x00,
                               "signal::completed", hide_when_done,
                               priv->lowlight,
                               NULL);
    }
}
